// src/Database.hh
// This file is part of libpbe; see http://decimail.org
// (C) 2004 - 2017 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef libpbe_Database_hh
#define libpbe_Database_hh

#include "Exception.hh"
#include "FileDescriptor.hh"
#include "endian.hh"

#include <string>
#include <memory>
#include <atomic>

#include <libpq-fe.h>

#include <boost/lexical_cast.hpp>
#include <boost/type_traits.hpp>
#include <boost/noncopyable.hpp>
#include <boost/static_assert.hpp>
#include <boost/iterator/iterator_facade.hpp>

#include <string.h>


namespace pbe {

// private:  (FIXME use detail::)

// Various libpq types have custom deleters; use std::unique_ptr with a 
// custom deleter - but we need to wrap the libpq deleter function as a 
// functor.  (Isn't there a standard library feature to do this?)

template< typename T, void (*FUNCTION)(T) >
struct mk_functor {
  void operator()(T t) { FUNCTION(t); }
};

typedef std::unique_ptr<PGresult, mk_functor<PGresult*, PQclear>> pgresult_ptr;

// We need to map between C++ types and PostgreSQL types in query parameters.
// This enum lists all of the PostgreSQL types that we map to, plus one 
// sentinel: unknown_type indicates that no PostgreSQL type corresponds to the 
// supplied C++ type, and should result in a compile-time error.  Hmm, except 
// it's disabled; what's going on there?

enum typecode_t { /*unknown_type=0,*/
                  text_type, numeric_type, timestamptz_type, bytea_type,
                  float_type, double_type,
                  typecode_t_max };

// Now we use template specialisation to get the typecode_t corresponding to a C++ 
// type at compile time.  get_typecode<T>() is the typecode_t for type T.

// We need a wrapper for times, because time_t is just a typedef for an integer 
// type and so it can't be specialised separately.

struct Timestamp {
  time_t t;
};

// Use this wrapper for bytea:

struct Bytea {
  std::string str;
};

// The generic template is not implemented anywhere, so it will fail at link 
// time for these unsupported types.
template <typename T> inline typecode_t get_typecode(void);

template <> inline typecode_t get_typecode<std::string>(void) { return text_type; }
template <> inline typecode_t get_typecode<Bytea>(void)       { return bytea_type; }
template <> inline typecode_t get_typecode<int32_t>(void)     { return numeric_type; }
template <> inline typecode_t get_typecode<int64_t>(void)     { return numeric_type; }
template <> inline typecode_t get_typecode<Timestamp>(void)   { return timestamptz_type; }
template <> inline typecode_t get_typecode<float>(void)       { return float_type; }
template <> inline typecode_t get_typecode<double>(void)      { return double_type; }

// (We have distinct typecodes for float and double but not for int32/64; hmmmm.)

// public:

// class Database represents a connection to the database server.
// It's marked non-copyable because it's not clear what should be done with 
// the underlying libpq connection object in that case.
class Database: public boost::noncopyable {
public:

  // Connect to the database server using the supplied connection info.
  Database(std::string conninfo);

  // Disconnect on destruction.
  ~Database();

  // Get the file descriptor for the connction to the server.
  // This is useful if you want to use select() to wait for asynchronous 
  // notifications from the server.  You obviously mustn't read or write to 
  // it.
  const pbe::FileDescriptor& get_fd(void) const;

  // Return any asynchronous notifications received from the server.
  std::string get_any_notification(void);

private:
  // libpq data structure representing the connection.
  PGconn* pgconn;

  // Connection file descriptor.
  pbe::FileDescriptor conn_fd;

  // This is needed so that we can support nested transactions.
  // (PostgreSQL now has built-in support for this, so it should be reviewed.)
  bool transaction_in_progress;

  // Run arbitary queries
  void exec_sql(std::string cmd);

  friend class QueryCore;
  friend class Transaction;
};


// Exceptions thrown by the Database functions:

class DatabaseException: public Exception {
public:
  DatabaseException(PGconn* pgconn, std::string doing_what_):
    postgres_error(PQerrorMessage(pgconn)),
    doing_what(doing_what_) {}
  void report(std::ostream& s) const;
private:
  std::string postgres_error;
  std::string doing_what;
};

class QueryFailed: public DatabaseException {
public:
  QueryFailed(PGconn* pgconn, std::string query):
    DatabaseException(pgconn,"Executing query " + query) {}
};

class TypeError: public DatabaseException {
public:
  TypeError(PGconn* pgconn):
    DatabaseException(pgconn,"Type error") {}
};


// Transactions.  Typical usage:
// {
//   Transaction t(db);
//   ... run queries ...
//   t.commit();
// }
// If the transaction goes out of scope without commit() having been called, e.g. 
// because an exception was thrown and not caught, the transaction will be rolled 
// back.
// TODO: PostgreSQL now has some sort of built-in support for nested 
// transactions.
class Transaction: public boost::noncopyable {
public:
  Transaction(Database& database_);
  ~Transaction();
  void commit(void);
private:
  Database& database;
  bool nested;
  bool committed;
};
    

// Convert from the representation returned by libpq to a normal C++ type.

template <typename T>
T decode_pq_res(const char* data, int length);

template <>
inline int32_t decode_pq_res<int32_t>(const char* data, int length __attribute__((unused)))
{
  int32_t i;
  assert (length == sizeof(i));
  ::memcpy(&i, data, sizeof(i));
  return ntohl(i);  // Note that ntohl is defined to take a 32-bit value, not a long.
                    // Note also that it's defined to take an unsigned.
}

template <>
inline int64_t decode_pq_res<int64_t>(const char* data, int length __attribute__((unused)))
{
  int64_t i;
  ::memcpy(&i, data, sizeof(i));
  return ntoh64(i);  // Ditto signed vs unsigned.
}

template <>
inline Timestamp decode_pq_res<Timestamp>(const char* data, int length)
{
  // Timestamp values are returned by PostgreSQL as 64-bit microsecond
  // values.  946684800000000 is a magic number to convert to the Unix
  // 1970 epoch.
  int64_t t = decode_pq_res<int64_t>(data, length);
  assert (length == sizeof(t));
  return Timestamp{t/1000000 + 946684800};
}

template <>
inline std::string decode_pq_res<std::string>(const char* data, int length)
{
  return std::string(data,length);
}

template <>
inline Bytea decode_pq_res<Bytea>(const char* data, int length)
{
  return Bytea{ std::string(data,length) };
}

template <>
inline float decode_pq_res<float>(const char* data, int length)
{
  int32_t i = decode_pq_res<int32_t>(data,length);
  float f;
  assert (length == sizeof(f));
  memcpy(&f, &i, sizeof(f));
  return f;
}

template <>
inline double decode_pq_res<double>(const char* data, int length)
{
  int64_t i = decode_pq_res<int64_t>(data,length);
  double d;
  assert (length == sizeof(d));
  memcpy(&d, &i, sizeof(d));
  return d;
}


// Result of a SELECT query:
class Result {
public:
  // Construct from the result object from libpq
  Result(pgresult_ptr&&);

  // For insert, update and delete statements the result tells us how many 
  // rows were inserted, updated or deleted, using the following:
  int get_rows_changed(void) const;

  // Everything below here is for getting at the table returned for a select 
  // statement:

  // The size of the table
  const int rows;
  const int cols;

  // Map between column names and numbers
  int column(std::string name) const ;
  std::string column_name(int pos) const;

  // Get data
  // The pointer returned by rawget is valid for as long as the Result object exists.
  const char* rawget(int row, int col) const;

  int getlength(int row, int col) const;

  template <typename T>
  T get_nocheck(int row, int col) const {
    return decode_pq_res<T>(rawget(row,col),getlength(row,col));
  }

  template <typename T>
  T get(int row, int col) const {
    check_column_type<T>(col);
    return get_nocheck<T>(row,col);
  }

  template <typename T>
  T get(int row, std::string colname) const {
    return get<T>(row, column(colname));
  }

  bool is_null(int row, int col) const;

  bool is_null(int row, std::string colname) const {
    return is_null(row, column(colname));
  }

  // Column types
  typecode_t column_typecode(int col) const;

  template <typename T>
  void check_column_type(int col) const {
    if (column_typecode(col)!=get_typecode<T>()) {
      throw StrException("type error for column "
            +boost::lexical_cast<std::string>(col)+", expecting typecode "
            +boost::lexical_cast<std::string>(get_typecode<T>())+" but got typecode "
            +boost::lexical_cast<std::string>(column_typecode(col)));
    }
  }

private:
  pgresult_ptr res;
};


// Alternative result for a query that generates a single column.
// Values can be accessed using (), e.g. r(i).
template <typename T>
class ColumnResult: public Result {
public:
  ColumnResult(Result&& r): Result(std::move(r)) {
    if (cols!=1) {
      throw "Single column expected";
    }
    if (rows>0) {
      check_column_type<T>(0);
    }
  }

  T operator ()(int row) const { return get<T>(row,0); }

  bool is_null (int row) const { return Result::is_null(row,0); }

  class const_iterator:
    public boost::iterator_facade<const_iterator,const T,boost::bidirectional_traversal_tag,T> {
  public:
    const_iterator():
      res(NULL), row(0) {}  // Does that work?
    const_iterator(const const_iterator& other):
      res(other.res), row(other.row) {}
  private:
    const ColumnResult& res;
    int row;
    friend class ColumnResult;
    const_iterator(const ColumnResult& res_, int row_):
      res(res_), row(row_) {}
    friend class boost::iterator_core_access;
    void increment() { ++row; }
    void decrement() { --row; }
    void advance(int n) { row+=n; }
    int distance_to(const_iterator other) { return other.row-row; }
    bool equal(const const_iterator& other) const {
      return row==other.row && &res==&(other.res);
    }
    const T dereference() const {
      return res(row);
    }
  };
  const_iterator begin() const {
    return const_iterator(*this,0);
  }
  const_iterator end() const {
    return const_iterator(*this,rows);
  }
};


// Alternative result for a query that generates a single value
// (e.g. a "count(*)" query).
// This is convertable to the type of that value.
template <typename T>
class SingletonResult: public Result {
public:
  SingletonResult(Result&& r): Result(std::move(r)) {
    if (rows!=1 || cols!=1) {
      throw "Singleton expected";
    }
    check_column_type<T>(0);
  }
  operator T () const { return get<T>(0,0); }
  bool is_null () const { return Result::is_null(0,0); }
};


// Alternative result for a query that generates a zero or one values.
// This is convertable to the type of that value.
template <typename T>
class OptResult: public Result {
public:
  OptResult(Result&& r): Result(std::move(r)) {
    if (rows>1 || cols!=1) {
      throw "Zero or one values expected";
    }
    if (rows==1) {
      check_column_type<T>(0);
    }
  }
  operator T () { return get<T>(0,0); }
  bool is_null () { return Result::is_null(0,0); }
  bool empty() { return rows==0; }
};


// private:

// Generate names for prepared statements
class statement_name_t: public std::string {
public:
  statement_name_t(void):
    std::string("stmt_"+boost::lexical_cast<std::string>(counter++))
  {}
private:
  static std::atomic<int> counter;
};


// Most of the work of Query is delagated to QueryCore which is not a template 
// class and so need not be coded inline.

class QueryCore {
public:
  QueryCore(Database& database_, std::string querystr_, int nparams,
            const typecode_t* argtypecodes, const int* formats);

  template <size_t NPARAMS>
  QueryCore(Database& database_, std::string querystr_,
            const std::array<typecode_t,NPARAMS>& argtypecodes,
            const std::array<int,NPARAMS>& formats):
    QueryCore(database_, querystr_, NPARAMS, argtypecodes.data(), formats.data())
  {}

  ~QueryCore();
  Result operator()(const char* enc_args[], const int* param_lengths);
  Result runonce(const char* enc_args[], const int* param_lengths);

private:
  Database& database;
  const std::string querystr;
  const bool params_ok;
  const statement_name_t statement_name;
  int nparams;
  const int* param_formats;
  std::unique_ptr<Oid[]> argoids;
  bool prepared;
  void prepare(void);
};


// Convert from C++ types to the representation passed to libpq.
// Types are either binary or character.  Binary is used for numeric types and 
// character is used for text types.  Binary types must be in network byte 
// order.
// This uses template specialisation.
// These functions return pointers; the pointers must remain valid until the 
// query completes, and the args to the query are sure to last longer than 
// that, so it's fine to return a pointer to the arg.  If the transform 
// is not-trivial it's not obvious how to deal with temporaries, but we 
// can cope by modifying the parameters in place; this is OK as long as 
// we don't need more space.

template <typename T>
inline const char* encode_pq_arg(T& t);  // Deliberately not defined.

template <>
inline const char* encode_pq_arg<int32_t>(int32_t& i)
{
  i = htonl(i);
  return reinterpret_cast<const char*>(&i);
}

template <>
inline const char* encode_pq_arg<int64_t>(int64_t& i)
{
  i = hton64(i);
  return reinterpret_cast<const char*>(&i);
}

template <>
inline const char* encode_pq_arg<Timestamp>(Timestamp& t)
{
  t.t = (t.t-946684800)*1000000;
  return encode_pq_arg<int64_t>(t.t);
}

template <>
inline const char* encode_pq_arg<std::string>(std::string& s) {
  return s.c_str();
}

template <>
inline const char* encode_pq_arg<Bytea>(Bytea& b) {
  return b.str.c_str();
}

template <>
inline const char* encode_pq_arg<float>(float& f) {
  uint32_t i;
  memcpy(&i, &f, sizeof(f));
  i = htonl(i);
  memcpy(&f, &i, sizeof(i));
  return reinterpret_cast<const char*>(&f);
}

template <>
inline const char* encode_pq_arg<double>(double& d) {
  uint64_t i;
  memcpy(&i, &d, sizeof(d));
  i = hton64(i);
  memcpy(&d, &i, sizeof(d));
  return reinterpret_cast<const char*>(&d);
}


// We need to know the size of the encoded data.
// This is sizeof(T) in most cases (we don't change the size while encoding 
// e.g. ints and floats); for strings it is the length of the string - though 
// it is ignored when the param_format is text.

template <typename T>
inline int get_param_length(const T&) { return sizeof(T); }

template <>
inline int get_param_length<std::string>(const std::string& s) { return s.length(); }

template <>
inline int get_param_length<Bytea>(const Bytea& b) { return b.str.length(); }


// Text/binary format; string is text (0) and others are binary (1).
// (Note Bytea is now binary, so length above is definitely needed.)

template <typename T>
struct get_format {
  static const int value = 1;
};

template <>
struct get_format<std::string> {
  static const int value = 0;
};

template <>
struct get_format<Bytea> {
  static const int value = 1;
};


// This base class is used to form the parameters passed to Query, below, into arrays 
// that can be passed to QueryCore.

template <typename... PARAMS>
struct QueryBase
{
  std::array<typecode_t, sizeof...(PARAMS)> typecodes;
  std::array<int,        sizeof...(PARAMS)> formats;

  QueryBase():
    typecodes{ get_typecode<PARAMS>()... },
    formats{ (get_format<PARAMS>::value)... }
  {}
};


// Prepare and run a query.  Typical usage:
// Query<std::string,int> insert_item(db,"insert into items(name,qty) values ($1,$2)");
// insert_item("nut",100);
// insert_item("bolt",102);

template <typename... PARAMS>
class Query:
  private QueryBase<PARAMS...>
{
public:

  // Create a query.  SQL is supplied in querystr with parameters indicated by 
  // $1, $2 etc.  C++ types of these parameters are the template paramters.
  Query(Database& database, std::string querystr):
    core(database, querystr,
         QueryBase<PARAMS...>::typecodes,
         QueryBase<PARAMS...>::formats)
  {}

  // Run a query.  It is prepared the first time that it is run.
  Result operator()(PARAMS... params...)
  {
    const char* arg_cs[sizeof...(PARAMS)] = { encode_pq_arg(params)... };
    int param_lengths[sizeof...(PARAMS)]  = { get_param_length(params)... };
    return core(arg_cs, param_lengths);
  }

  // Run a query immediately, without preparation.  Use this for queries that will only be 
  // run once.
  Result runonce(PARAMS... params...)
  {
    const char* arg_cs[sizeof...(PARAMS)] = { encode_pq_arg(params)... };
    int param_lengths[sizeof...(PARAMS)]  = { get_param_length(params)... };
    return core.runonce(arg_cs, param_lengths);
  }


private:
  QueryCore core;
};


// A query that returns a ColumnResult.
template <typename RT, typename... PARAMS>
class ColumnQuery: public Query< PARAMS...> {
public:
  typedef ColumnResult<RT> result_t;

  ColumnQuery(Database& database, std::string querystr):
    Query<PARAMS...>(database,querystr)
  {}

  result_t operator()(PARAMS... params...) {
    return Query<PARAMS...>::operator()(params...);
  }

  result_t runonce(PARAMS... params...) {
    return Query<PARAMS...>::runonce(params...);
  }
};


// A query that returns a SingeltonResult.  In this case there's no need to 
// delcare the result object, thanks to implicit type conversion.  For 
// example:  int a = query(param);
template <typename RT, typename... PARAMS>
class SingletonQuery: public Query< PARAMS...> {
public:
  typedef SingletonResult<RT> result_t;

  SingletonQuery(Database& database, std::string querystr):
    Query<PARAMS...>(database,querystr)
  {}

  result_t operator()(PARAMS... params...) {
    return Query<PARAMS...>::operator()(params...);
  }

  result_t runonce(PARAMS... params...) {
    return Query<PARAMS...>::runonce(params...);
  }
};


// A query that returns a OptResult.
template <typename RT, typename... PARAMS>
class OptQuery: public Query<PARAMS...> {
public:
  typedef OptResult<RT> result_t;

  OptQuery(Database& database, std::string querystr):
    Query<PARAMS...>(database,querystr)
  {}

  result_t operator()(PARAMS... params...) {
    return Query<PARAMS...>::operator()(params...);
  }

  result_t runonce(PARAMS... params...) {
    return Query<PARAMS...>::runonce(params...);
  }
};


};

#endif
