// src/Date.hh
// This file is part of libpbe; see http://decimail.org
// (C) 2006-2014 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef libpbe_Date_hh
#define libpbe_Date_hh

#include "format.hh"

#include <iostream>

#include <time.h>

#include <stdint.h>


namespace pbe {


  static int days_in_month_[] = { 31, 0, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
  static int days_to_start_of_month_[] = {
    0,
    31,
    31+28,
    31+28+31,
    31+28+31+30,
    31+28+31+30+31,
    31+28+31+30+31+30,
    31+28+31+30+31+30+31,
    31+28+31+30+31+30+31+31,
    31+28+31+30+31+30+31+31+30,
    31+28+31+30+31+30+31+31+30+31,
    31+28+31+30+31+30+31+31+30+31+30 };

  struct Date {

    uint16_t year;
    uint8_t month;
    uint8_t day;

    Date() {}

    Date(int year_, int month_, int day_):
      year(year_), month(month_), day(day_) {}

    Date(time_t t) {
      // This is Howard Hinnant's algorithm; see http://howardhinnant.github.io/date_algorithms.html
      // I've removed the code that copes with BC dates.
      const int z = t / 86400 + 719468;
      const int era = z / 146097;
      const unsigned doe = static_cast<unsigned>(z - era * 146097);
      const unsigned yoe = (doe - doe/1460 + doe/36524 - doe/146096) / 365;
      const int y = static_cast<int>(yoe) + era * 400;
      const unsigned doy = doe - (365*yoe + yoe/4 - yoe/100);
      const unsigned mp = (5*doy + 2)/153;
      day = doy - (153*mp+2)/5 + 1;
      month = mp + (mp < 10 ? 3 : -9);
      year = y + (month <= 2);
    }

    bool operator<(const Date& rhs) const {
      return (year<rhs.year)
          || (year==rhs.year && month<rhs.month)
          || (year==rhs.year && month==rhs.month && day<rhs.day);
    }

    bool operator==(const Date& rhs) const {
      return (year==rhs.year && month==rhs.month && day==rhs.day);
    }

    bool operator!=(const Date& rhs) const {
      return !operator==(rhs);
    }

    bool operator>(const Date& rhs) const {
      return rhs < *this;
    }

    bool operator<=(const Date& rhs) const {
      return ! (*this > rhs);
    }

    bool operator>=(const Date& rhs) const {
      return ! (*this < rhs);
    }

    bool is_leap_year(void) const {
      return !(year%4) && ((year%100) || !(year%400));
    }

    int days_in_month(void) const {
      if (month==2) {
        if (is_leap_year()) {
          return 29;
        } else {
          return 28;
        }
      } else {
        return days_in_month_[month-1];
      }
    }

    int days_since_1970_1_1() const
    {
      // Days from 1st Jan 1970 to this date.
      int years_since_1970 = year-1970;
      int leap_years_since_1970 = (year-1969)/4;  // not including this year
      int days_to_start_of_year = 365 * years_since_1970 + leap_years_since_1970;
      int days = days_to_start_of_year
               + days_to_start_of_month_[month-1]
               + ((month>2 && is_leap_year()) ? 1 : 0)
               + day-1;
      return days;
    }

    time_t to_time_t() const {
      // Seconds from 1st Jan 1970 to midnight at start of this date.
      return days_since_1970_1_1() * 24*60*60;
    }


    int day_of_week(void) const  // 1 = Monday, 7 = Sunday
    {
      // 1st Jan 1970 was a Thursday.
      return ((days_since_1970_1_1() + 3) % 7) + 1;
    }

    Date& operator++() {
      if (day<days_in_month()) {
        ++day;
      } else {
        day = 1;
        if (month<12) {
          ++month;
        } else {
          month=1;
          ++year;
        }
      }
      return *this;
    }

    Date& operator--() {
      if (day>1) {
        --day;
      } else {
        if (month>1) {
          --month;
        } else {
          month=12;
          --year;
        }
        day = days_in_month();
      }
      return *this;
    }

  };


  // FIXME this doesn't insert leading 0s, but format() below does.
  inline std::ostream& operator<<(std::ostream& strm, Date d) {
    strm << d.year << "-" << static_cast<int>(d.month) << "-" << static_cast<int>(d.day);
    return strm;
  }

  inline std::string format(const Date& d) {
    return format("%04d-%02d-%02d",d.year, d.month, d.day);
  }

}


#endif
