#ifndef libpbe_alphanum_compare_hh
#define libpbe_alphanum_compare_hh

// include/alphanum_compare.hh
// This file is part of libpbe; see http://svn.chezphil.org/libpbe/
// (C) 2012 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <cctype>

// alphanum_compare is like std::lexicographical_compare except that it 
// treats sequences of numbers embedded in the input differently.
// So for example, it sorts foo2 as less than foo10.


namespace pbe {


template <typename ITER1, typename ITER2>
bool alphanum_compare(ITER1 first1, ITER1 last1,
                      ITER2 first2, ITER2 last2 )
{
  // Is the range [first1,last1) less than the range [first2,last2) ?
  
  while (first1 != last1 && first2 != last2) {
    char c1 = *first1;
    char c2 = *first2;
    ++first1;
    ++first2;

    // If neither character is a digit, just compare them.
    // If one is a digit and the other isn't, again we can just compare them.
    if (!std::isdigit(c1) || !std::isdigit(c2)) {
      if (c1 < c2) return true;
      if (c1 > c2) return false;
      continue;
    }

    // Both characters are digits.
    // Longer numbers are larger (? except when they have leading zeros)
    std::size_t l1 = 1;
    for (ITER1 i1 = first1; i1 != last1 && std::isdigit(*i1); ++i1) ++l1;
    std::size_t l2 = 1;
    for (ITER1 i2 = first2; i2 != last2 && std::isdigit(*i2); ++i2) ++l2;
    if (l1 < l2) return true;
    if (l1 > l2) return false;

    // Both numbers have the same number of digits; compare them numerically.
    for (std::size_t ix=0; ix<l1; ++ix) {
      if (c1 < c2) return true;
      if (c1 > c2) return false;
      c1 = *first1;
      c2 = *first2;
      ++first1;
      ++first2;
    }
  }

  // We've reached the end of one or both inputs.
  // If we've reached the end of both, they are equal - return false.
  // If we've reached the end of 1 but not 2, then 1 is less - return true.
  // If we've reached the end of 2 but not 1, then 2 is less - return false.
  // => return true iff we've not reached the end of 2.
  return first2 != last2;
}


// Helper to compare entire sequences:

template <typename CONT1, typename CONT2>
bool alphanum_compare(const CONT1& c1, const CONT2& c2)
{
  return alphanum_compare(c1.begin(),c1.end(), c2.begin(),c2.end());
}


};


#endif

