// include/transaction.hh
// This file is part of libpbe; see http://svn.chezphil.org/libpbe/
// (C) 2013 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef libpbe_transaction_hh
#define libpbe_transaction_hh


// Provide transactions for additions to containers.
// Usage:
//
//   typedef std::vector<int> vec;
//   vec v;
//   transaction<vec> t(v);
//   v.push_back(f());
//   v.push_back(g());
//   t.commit();
//
// If f() or g() throws, t is destroyed without committing; this causes the 
// container to be rolled back to its size when t was created.
//
// Note that only changes adding items beyond the original end of the container 
// can be rolled back.
//
// The container can be any that provides size() and erase(i,j), i.e. std::vector<T>, 
// std::list<T>, std::deque<T>, etc.  (Note that resize() could be used for the 
// rollback, but this requires that T be default-constructible for the unused 
// resize-larger case.)
//
// Because rollback happens during exception processing, it should not throw - 
// so T::~T() should not throw.

#include <boost/noncopyable.hpp>


namespace pbe {


template <typename T>
class transaction: boost::noncopyable {
  T& cont;
  const size_t orig_size;
  bool committed;
  void rollback() {
    typename T::iterator i = cont.end();
    std::advance(i,orig_size-cont.size());
    cont.erase(i,cont.end());
  }
public:
  transaction(T& cont_):
    cont(cont_),
    orig_size(cont.size()),
    committed(false)
  {}
  ~transaction() {
    if (!committed) rollback();
  }
  void commit() {
    committed = true;
  }
};


};  // namespace pbe


#endif

