// src/bstgmt.cc
// This file is part of libpbe; see http://svn.chezphil.org/libpbe/
// (C) 2017 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include "bstgmt.hh"

#include <elemsof.hh>
#include <Date.hh>
#include <flat_multimap.hh>

#include <algorithm>


namespace pbe {

// BST dates from http://www.npl.co.uk/educate-explore/what-is-time/archive-of-summer-time-dates-1916-2006
// Note that odd things happened in 1970-71.  So I'll start the table in 1972.

struct raw_bst_period { int year; int start_month; int start_day; int end_month; int end_day; };
const raw_bst_period raw_bst_periods[] = {
       // Start and end dates are both the Sunday.
       { 1972, 3, 19, 10, 29 },
       { 1973, 3, 18, 10, 28 },
       { 1974, 3, 17, 10, 27 },
       { 1975, 3, 16, 10, 26 },
       { 1976, 3, 21, 10, 24 },
       { 1977, 3, 20, 10, 23 },
       { 1978, 3, 19, 10, 29 },
       { 1979, 3, 18, 10, 28 },
       { 1980, 3, 16, 10, 26 },
       { 1981, 3, 29, 10, 25 },
       { 1982, 3, 28, 10, 24 },
       { 1983, 3, 27, 10, 23 },
       { 1984, 3, 25, 10, 28 },
       { 1985, 3, 31, 10, 27 },
       { 1986, 3, 30, 10, 26 },
       { 1987, 3, 29, 10, 25 },
       { 1988, 3, 27, 10, 23 },
       { 1989, 3, 26, 10, 29 },
       { 1990, 3, 25, 10, 28 },
       { 1991, 3, 31, 10, 27 },
       { 1992, 3, 29, 10, 25 },
       { 1993, 3, 28, 10, 24 },
       { 1994, 3, 27, 10, 23 },
       { 1995, 3, 26, 10, 22 },
       { 1996, 3, 31, 10, 27 },
       { 1997, 3, 30, 10, 26 },
       { 1998, 3, 29, 10, 25 },
       { 1999, 3, 28, 10, 31 },
       { 2000, 3, 26, 10, 29 },
       { 2001, 3, 25, 10, 28 },
       { 2002, 3, 31, 10, 27 },
       { 2003, 3, 30, 10, 26 },
       { 2004, 3, 28, 10, 31 },
       { 2005, 3, 27, 10, 30 },
       { 2006, 3, 26, 10, 29 },
       { 2007, 3, 25, 10, 28 },
       { 2008, 3, 30, 10, 26 },
       { 2009, 3, 29, 10, 25 },
       { 2010, 3, 28, 10, 31 },
       { 2011, 3, 27, 10, 30 },
       { 2012, 3, 25, 10, 28 },
       { 2013, 3, 31, 10, 27 },
       { 2014, 3, 30, 10, 26 },
       { 2015, 3, 29, 10, 25 },
       { 2016, 3, 27, 10, 30 },
       { 2017, 3, 26, 10, 29 },
       { 2018, 3, 25, 10, 28 },
       { 2019, 3, 31, 10, 27 }
};


typedef pbe::flat_multimap<time_t,time_t> bst_periods_t;

static bst_periods_t build_bst_periods()
{
  bst_periods_t bst_periods;
  bst_periods.reserve(pbe::elemsof(raw_bst_periods));
  for (size_t i = 0; i < pbe::elemsof(raw_bst_periods); ++i) {
    const raw_bst_period& r = raw_bst_periods[i];
    pbe::Date start_date(r.year,r.start_month,r.start_day);
    pbe::Date end_date(r.year,r.end_month,r.end_day);
    // Start time is 1AM GMT == 2AM BST, end time is at 2AM BST = 1AM GMT
    // pbe::Date::time_t() returns seconds to midnight at the start of the 
    // date UTC, without any time zone considerations.
    time_t s = start_date.to_time_t() + 3600;
    time_t e = end_date.to_time_t() + 3600;
    bst_periods.insert(std::make_pair(s,e));
  }
  bst_periods.sort();
  return bst_periods;
}


static const bst_periods_t& the_bst_periods()
{
  static const bst_periods_t the = build_bst_periods();
  return the;
}


bool is_in_bst(time_t t)
{
  const bst_periods_t& bst_periods = the_bst_periods();
  bst_periods_t::const_iterator i = bst_periods.upper_bound(t);
  // i refers to the first period that begins after t
  if (i == bst_periods.begin()) return false;  // Before March 1972
  --i;
  // i refers to the last period that begins before or at t
  return t < i->second;
}


time_t apply_offset_for_bst(time_t t)
{
  if (is_in_bst(t)) {
    t += 3600;
  }
  return t;
}


time_t remove_offset_for_bst(time_t t)
{
  // At the end of BST, a time between 01:00 and 01:59 is ambiguous. 
  // 01:30 GMT == 02:30 BST.  is_in_bst(01:30 GMT) will return false. 
  // So we don't apply the offset and imply that 01:30 is GMT i.e. 
  // after the change.
  if (is_in_bst(t)) {
    t -= 3600;
  }
  return t;
}


};

